var facade = require("facade");
var Objects = require("convenience").Objects;
var _ = require("lodash");
var on = require("on")(plugin);
var API = require("../cattura.web.api.APIPlugin/apiHelper")(plugin);

on({
    WebServerEvent: function(event)
    {
        var webServer = event.getWebServer();
        var environment = facade.retrieveProxy("EnvironmentProxy");

        var File = java.io.File;
        var FileWrapper = Packages.cattura.utils.io.wrappers.FileWrapper;
        var EventSourceHandler = org.webbitserver.EventSourceHandler;
        var EventSourceMessage = org.webbitserver.EventSourceMessage;
        var Permissions = Packages.cattura.agents.configuration.model.properties.Permissions;
        var RequestResponseHelper = Packages.cattura.web.server.vos.RequestResponseHelper;
        var HttpHandler = org.webbitserver.HttpHandler;

        var LOGGER = plugin.info().logger;
        var STORAGE = environment.getStorageDirPath().getCanonicalFile();

        var invalidRootDirectories = [
            STORAGE, // Root storage directory
            environment.getCaptureSessionPath().getCanonicalFile(), // Capture session directory
            environment.getTemplatesDirPath().getCanonicalFile() // Templates directory
        ];

        var capture = facade.retrieveProxy("MediaProcessingProxy");

        function path (root)
        {
            if (!root)
            {
                throw "Empty path not allowed!";
            }

            var thisPath = new File(STORAGE.getPath() + File.separator + root).getCanonicalFile();

            if (!thisPath.getPath().startsWith(STORAGE.getPath()))
            {
                throw "Invalid path [" + thisPath + "]";
            }

            for (var i = 0; i < invalidRootDirectories.length; i++)
            {
                var invalidRootDirectory = invalidRootDirectories[i];
                if (thisPath.equals(invalidRootDirectory))
                {
                    throw "Can't put a graph at root directory [" + thisPath + "]";
                }
            }

            return thisPath.getPath();
        }

        function builder ()
        {
            return capture.getGraphBuilder();
        }

        function graphAt (root)
        {
            return builder().get(new File(path(root)));
        }

        function parseGraph (graph)
        {
             return JSON.parse(graph.asString());
        }

        function findNodeInGraph (graph, nodeID)
        {
            for (var i = 0; i < graph.nodes.length; i++)
            {
                var node = graph.nodes[i];
                if (node.nodeID == nodeID)
                {
                    return node;
                }
            }

            return null;
        }

        function findSavedGraphs (prefix)
        {
            var paths = [];
            prefix = prefix || "";

            (function search(directory) {
                _(directory.listFiles()).each(function(file) {
                    if (file.isDirectory())
                    {
                        search(file);
                    }
                    else if(file.name == "graph.json")
                    {
                        paths.push(Objects.toJS(file.parent).substr(STORAGE.absolutePath.length() + 1));
                        return; // we found graph.json here, no need to keep looking
                    }
                });
            })(new File(STORAGE.absolutePath + File.separator + prefix));

            return paths;
        }

        /*
         * A "SessionStatus" object listens to status callbacks from a capture session.
         * It also creates and registers an EventSource on the API.
         */
        sessionStatus = (function() {
            var statii = {};

            return function sessionStatus (root)
            {
                /*
                 * To ensure we only have one instance of this per graph root.
                 */
                if(!(root in statii))
                {
                    statii[root] = (function(root) {
                        var messages = [];
                        var nodes = {};
                        var eventsource = null;

                        /*
                         * Status callback for the graph.
                         */
                        graphAt(root).onStatusUpdate(function(status) {
                            sessionStatus(root).update(status);
                        });

                        return {
                            nodes: function ()
                            {
                                return _(nodes).clone();
                            },
                            update: function (status)
                            {
                                messages.push(status);

                                if (status.state !== null)
                                {
                                    nodes[status.nodeID] = Objects.toJS(status.state.toString());
                                    this.eventSource().send(status);

                                    if (status.nodeID === 0 && status.state.toString() == "TERMINATED")
                                    {
                                        this.eventSource().close();
                                    }
                                }
                            },
                            messages: function ()
                            {
                                return _(messages).without().value();
                            },
                            eventSource: function ()
                            {
                                if (eventsource === null)
                                {
                                    eventsource = (function() {
                                        var connections = [];

                                        return {
                                            uri: function ()
                                            {
                                                return "sessionStatus" + root;
                                            },
                                            send: function (status)
                                            {
                                                _(connections).each(function(connection) {
                                                    connection.send(new EventSourceMessage().event("status").data(JSON.stringify({
                                                        state: Objects.toJS(status.state.toString()),
                                                        nodeID: Objects.toJS(status.nodeID),
                                                        pluginID: Objects.toJS(status.pluginID),
                                                        reason: Objects.toJS((status.reason))
                                                    })));
                                                });
                                            },
                                            close: function ()
                                            {
                                                _(connections).each(function(connection) {
                                                    connection.close();
                                                });
                                            },
                                            handler: new EventSourceHandler({
                                                onOpen: function (connection)
                                                {
                                                    LOGGER.info(connection);
                                                    connections.push(connection);
                                                },
                                                onClose: function (connection)
                                                {
                                                    connections = _(connections).without(connection).value();
                                                }
                                            })
                                        };
                                    })();

                                    webServer.add(eventsource.uri(), eventsource.handler);
                                }

                                return eventsource;
                            }
                        };
                    })(root);
                }

                return statii[root];
            };
        })();

        function listFilesForNode (node)
        {
            var outputJSON = new File(node.workingDir(), "output.json");
            try
            {
                outputJSON = outputJSON.exists() ? JSON.parse(new FileWrapper(outputJSON).read()) : {};
            }
            catch (e)
            {
                outputJSON = {};
            }

            var outputs = _(node.workingDir().listFiles()).map(function(file) {
                var output = {
                    name: Objects.toJS(file.name),
                    size: file.length(),
                    type: Objects.toJS(new FileWrapper(file).type())
                };

                if (output.name in outputJSON)
                {
                    output.element = outputJSON[output.name];
                    delete outputJSON[output.name];
                }

                return output;
            }).value();

            for (var key in outputJSON)
            {
                if (outputJSON.hasOwnProperty(key))
                {
                    outputs.push({
                        name: key,
                        size: 0,
                        type: "unknown",
                        element: outputJSON[key]
                    });
                }
            }

            return outputs;
        }

        function getDownloadURLForFile (request, node, fileName)
        {
            return new RequestResponseHelper(request, null).getRelativeRoot() + "storage" + 
                (node.workingDir().getAbsolutePath().substring(STORAGE.getAbsolutePath().length()) + "/" + 
                 fileName).replace(/\/\//gi, "/");
        }

        /*
         * The Graph Builder API plugin provides 3 APIs. The GraphBuilder API is used for the creation and
         * management of capture session graphs. The SessionStatus API is used for getting the status of
         * and recieving updates from a currently running capture session. The SessionOutput API provides
         * access to the output of a capture session.
         */
        API({

           /*
            * The GraphBuilder API provides mechanisms to create, modify, and run capture session graphs.
            * It uses none of the legacy Hardware Agent classes such as MediaPackageVO or LectureTemplateVO,
            * nor does it "know about" the specific kinds of metadata (such as presentation title or speaker data)
            * that the Hardware Agent automatically adds.
            *
            * Wherever possible, operations that operate on entire graphs will return the JSON text of the graph.
            * Operations that operate on specific nodes of the graph will return the JSON text of that node. Exceptions
            * include the Copy function, which instead returns a new graph, and the StartCapture/StopCapture functions.
            */
            graphBuilder: {
                title: "Graph Builder",
                description: "Advanced methods for constructing and managing capture session graphs.",
                version: 1,
                methods: {
                    /*
                     * Friendly wrapper around the base API
                     */
                    apiWrapper: {
                        description: "Returns a friendly wrapper for the graph builder API.",
                        verb: "get",
                        endpoint: "graphBuilder/graph.js",
                        handler: new HttpHandler({
                            handleHttpRequest: function(request, response, control) {
                                response.header('Content-Type', 'application/javascript').content(
                                    plugin.resource("htdocs", "graph.js").read()
                                ).end();
                            }
                        })
                    },
                    listGraphs: {
                        description: "Return a list of all saved graphs.",
                        verb: "get",
                        endpoint: "graphBuilder",
                        callback: function(prefix) {
                            return findSavedGraphs(prefix);
                        }
                    },

                    /*
                     * Operations on an entire graph take a root path of the graph in the URL.
                     *
                     * Graphs created through the API are constrained to the DefaultPaths.DEFAULT_STORAGE_PATH, so a request
                     * to http://192.168.1.111graphBuilder/foo/bar/baz actually goes to {DEFAULT_STORAGE_PATH}/foo/bar/baz.
                     */
                    save: {
                        description: "Saves the current graph in memory to the filesystem.",
                        endpoint: "graphBuilder{/root}/save",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the existing graph's directory structure"},
                        },
                        callback: function(root) {
                            return parseGraph(graphAt(root).save());
                        }
                    },
                    copy: {
                        description: "Create a copy of an existing graph with a new root.",
                        endpoint: "graphBuilder{/root}/copy",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the existing graph's directory structure"},
                            copyTo: {type: "string", description: "the root of the copy's graph directory structure"}
                        },
                        callback: function(root, copyTo) {
                            return parseGraph(graphAt(root).copy(new File(copyTo)));
                        }
                    },
                    createNode: {
                        description: "Construct a processor node and attach it to a graph.",
                        endpoint: "graphBuilder{/root}/createNode",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            pluginID: {type: "string", description: "ID of plugin to construct"},
                            config: {type: "json", description: "if specified, a configuration for the plugin"},
                            parentNodeIDs: {type: "integer list", description: "if specified, parent nodes for the node"}
                        },
                        callback: function(root, pluginID, config, parentNodeIDs) {
                            var graph = graphAt(root);
                            var node = graph.add(pluginID);

                            if (!Objects.isEmpty(config))
                            {
                                node.configure(JSON.parse(config));
                            }

                            if (!Objects.isEmpty(parentNodeIDs))
                            {
                                node['linkParents(java.lang.Iterable)'](_(parentNodeIDs).map(function(val) {
                                    return graph.node(val).get();
                                })).value();
                            }

                            return findNodeInGraph(parseGraph(graphAt(root)), node.id());
                        }
                    },
                    startCapture: {
                        description: "Start a session at this graph location.",
                        endpoint: "graphBuilder{/root}/startCapture",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                        },
                        callback: function(root) {
                            var graph = graphAt(root);

                            /*
                             * Set up EventSource for status listeners
                             */
                            var eventsource = sessionStatus(root).eventSource();

                            /*
                             * Start capture
                             */
                            graph.startCapture();

                            return eventsource.uri();
                        }
                    },
                    stopCapture: {
                        description: "Stops a session at this graph location, if there is one.",
                        endpoint: "graphBuilder{/root}/stopCapture",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                        },
                        callback: function(root) {
                            var graph = graphAt(root);
                            graph.stopCapture();

                            return true;
                        }
                    },

                    /*
                     * Operations on individual nodes recieve the nodeID as a URL parameter.
                     */
                    linkNode: {
                        description: "Attach a node to one or more parents and/or children.",
                        endpoint: "graphBuilder{/root}/nodes/{nodeID}/link",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"},
                            parentNodeIDs: {type: "integer list", description: "list of parent node IDs to link the specified node to"},
                            childNodeIDs: {type: "integer list", description: "list of child node IDs to link the specified node to"}
                        },
                        callback: function(root, nodeID, parentNodeIDs, childNodeIDs) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();
                            var parentNodes = _(parentNodeIDs).map(function(val) {
                                return graph.node(val).get();
                            }).value();
                            var childNodes = _(childNodeIDs).map(function(val) {
                                return graph.node(val).get();
                            }).value();

                            node['linkParents(java.lang.Iterable)'](parentNodes);
                            node['linkChildren(java.lang.Iterable)'](childNodes);

                            return findNodeInGraph(parseGraph(graphAt(root)), nodeID);
                        }
                    },
                    nodeOptions: {
                        description: "Retrieve the options for this node.",
                        endpoint: "graphBuilder{/root}/nodes/{nodeID}/options",
                        verb: "get",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"}
                        },
                        callback: function(root, nodeID) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();
                            return Objects.toJS(node.options(), true);
                        }
                    },
                    configureNode: {
                        description: "Configure the specified node.",
                        endpoint: "graphBuilder{/root}/nodes/{nodeID}/configure",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"},
                            config: {type: "json", description: "JSON configuration for the node"}
                        },
                        callback: function(root, nodeID, config) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();
                            node.configure(JSON.parse(config));

                            return findNodeInGraph(parseGraph(graphAt(root)), nodeID);
                        }
                    },
                    invokeService: {
                        description: "Invokes a service of this Processor.",
                        endpoint: "graphBuilder{/root}/nodes/{nodeID}/services/{service}",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"},
                            service: {type: "string", description: "Name of service to invoke"},
                            args: {type: "json", description: "JSON object of named arguments for the service"}
                        },
                        callback: function(root, nodeID, service, args) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();

                            return Objects.toJS(node.invokeServiceSynchronously(service, JSON.parse(args)), true);
                        }
                    },
                    getNode: {
                        description: "Get a single node from the graph.",
                        endpoint: "graphBuilder{/root}/nodes/{nodeID}",
                        require: Permissions.VIEW,
                        verb: "get",
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"}
                        },
                        callback: function(root, nodeID) {
                            return findNodeInGraph(parseGraph(graphAt(root)), nodeID);
                        }
                    },
                    removeNode: {
                        description: "Remove a node from the graph.",
                        endpoint: "graphBuilder{/root}/nodes/{nodeID}",
                        require: Permissions.VIEW,
                        verb: "delete",
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"}
                        },
                        callback: function(root, nodeID) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();
                            node.remove();

                            return parseGraph(graphAt(root));
                        }
                    },

                    /*
                     * These are placed down here because otherwise they would swallow up all other requests
                     */
                    getGraph: {
                        description: "Get the complete JSON text of a graph.",
                        endpoint: "graphBuilder{/root}",
                        require: Permissions.VIEW,
                        verb: "get",
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"}
                        },
                        callback: function(root) {
                            return parseGraph(graphAt(root));
                        }
                    },

                    make: {
                        description: "Create an new graph with a root node.",
                        endpoint: "graphBuilder{/root}",
                        require: Permissions.VIEW,
                        verb: "post",
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"}
                        },
                        callback: function(root) {
                            return parseGraph(graphAt(root).make());
                        }
                    }
                }
            },

            /*
             * The SessionStatus API provides information about currently running capture sessions.
             * It provides methods to query the Capture Agent for the status of each node.
             */
            sessionStatus: {
                title: "Session Status",
                description: "Allows you to query the state of a currently running capture session.",
                version: 1,
                methods: {
                    allNodes: {
                        description: "Get the current status of each node in a capture session.",
                        endpoint: "sessionStatus{/root}/nodes",
                        verb: "get",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"}
                        },
                        callback: function(root) {
                            return sessionStatus(root).nodes();
                        }
                    }
                }
            },

            /*
             * The SessionOutput API provides access to files generated from the capture session,
             * such as video, assets, and process logs.
             */
            sessionOutput: {
                title: "Session Output",
                description: "Provides access to files generated by the capture session.",
                version: 1,
                methods: {
                    forNode: {
                        description: "For this specific node, get all files that it produced.",
                        endpoint: "sessionOutput{/root}/nodes/{nodeID}",
                        verb: "get",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"}
                        },
                        callback: function(root, nodeID) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();

                            return listFilesForNode(node);
                        }
                    },
                    getFile: {
                        description: "For this specific node, get a specific file it produced.",
                        endpoint: "sessionOutput{/root}/nodes/{nodeID}{/fileName}",
                        verb: "get",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"},
                            nodeID: {type: "integer", description: "ID of the node to perform this operation on"},
                            fileName: {type: "string", description: "the name of the file to get"}
                        },
                        callback: function(root, nodeID, fileName) {
                            var graph = graphAt(root);
                            var node = graph.node(nodeID).get();

                            return new API.Response(function(request, response) {
                                response.status(302).header("Location", getDownloadURLForFile(request, node, fileName)).end();
                            });
                        }
                    },
                    forAllNodes: {
                        description: "For each node, get all files that it produced.",
                        endpoint: "sessionOutput{/root}",
                        verb: "get",
                        require: Permissions.VIEW,
                        params: {
                            root: {type: "string", description: "the root of the graph directory structure"}
                        },
                        callback: function(root) {
                            var graph = graphAt(root);
                            var nodeOutputs = {};
                            var nodes = Objects.toJS(graph.nodes());

                            for (var i = 0; i < nodes.length; i++)
                            {
                                var node = nodes[i];
                                nodeOutputs[node.id()] = listFilesForNode(node);
                            }

                            return nodeOutputs;
                        }
                    }
                }
            }
        });
    }
});